import pytest

from todo_example import spec, api as todo_api


@pytest.fixture
def api():
    return spec.TodoAPI.from_api(todo_api.TodoApp())


@pytest.mark.parametrize("impl", [spec.ViaAPI, spec.ViaUX, spec.ViaRPC])
def test_simple(api, impl):
    with api.use(impl):
        assert api.implementation_chooser.current.elements == (impl,)
        collection = api.create_collection(name="test")
        item = collection.create_item(name="buy ham")
        # assert collection.all()[0] == item

        try:
            assert not item.completed
        except LookupError:
            pass  # rpc
        item.completed = True

        try:
            assert item.completed
        except LookupError:
            pass  # rpc
        collection.clear_completed()
        assert collection.get_by(name="buy ham") is None



import sys
import os
import pkg_resources

sys.path.insert(0, os.path.abspath("../examples"))

__distribution = pkg_resources.get_distribution("Sentaku")

extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.doctest",
    "sphinx.ext.viewcode",
    "sphinx.ext.todo",
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]
source_suffix = ".rst"
master_doc = "index"

# General information about the project.
project = __distribution.project_name
copyright = "2016, Ronny Pfannschmidt"
author = "Ronny Pfannschmidt"

release = __distribution.version
version = ".".join(release.split(".")[:2])

exclude_patterns = []

pygments_style = "sphinx"
todo_include_todos = False

html_theme = "haiku"
html_static_path = ["_static"]

htmlhelp_basename = "sentakudoc"



import attr
import sentaku
from .ux import TodoUX

ViaAPI = sentaku.ImplementationName("API")
ViaUX = sentaku.ImplementationName("UX")
ViaRPC = sentaku.ImplementationName("RPC")


class TodoAPI(sentaku.ImplementationContext):
    """example description for a simple todo application"""

    @classmethod
    def from_api(cls, api):
        """
        create an application description for the todo app,
        that based on the api can use either tha api or the ux for interaction
        """
        ux = TodoUX(api)
        from .pseudorpc import PseudoRpc

        rpc = PseudoRpc(api)

        return cls({ViaAPI: api, ViaUX: ux, ViaRPC: rpc})

    create_collection = sentaku.ContextualMethod()


@attr.s
class TodoItem(sentaku.Element):
    """describing a todo list element"""

    name = attr.ib()
    completed = sentaku.ContextualProperty()


@TodoAPI.external_for(TodoAPI.create_collection, ViaAPI)
@TodoAPI.external_for(TodoAPI.create_collection, ViaUX)
def create_todo_collection(self, name):
    self.impl.create_item(name)
    return TodoCollection(self, name=name)


@attr.s
class TodoCollection(sentaku.Collection):
    """domain object describing a todo list"""

    name = attr.ib()
    create_item = sentaku.ContextualMethod()
    get_by = sentaku.ContextualMethod()
    clear_completed = sentaku.ContextualMethod()


@TodoAPI.external_for(TodoItem.completed.setter, ViaAPI)
@TodoAPI.external_for(TodoItem.completed.setter, ViaUX)
def completed(self, value):
    col = self.impl.get_by(self.parent.name)
    elem = col.get_by(self.name)
    elem.completed = value


@TodoAPI.external_for(TodoCollection.create_item, ViaAPI)
@TodoAPI.external_for(TodoCollection.create_item, ViaUX)
def create_item(self, name):
    collection = self.impl.get_by(self.name)
    elem = collection.create_item(name=name)
    assert elem
    return TodoItem(self, name=name)


@TodoAPI.external_for(TodoCollection.get_by, ViaAPI)
@TodoAPI.external_for(TodoCollection.get_by, ViaUX)
def get_by(self, name):
    collection = self.impl.get_by(self.name)
    elem = collection.get_by(name)
    if elem is not None:
        return TodoItem(self, name=name)


@TodoAPI.external_for(TodoCollection.clear_completed, ViaUX)
@TodoAPI.external_for(TodoCollection.clear_completed, ViaAPI)
def clear_completed(self):
    collection = self.impl.get_by(self.name)
    collection.clear_completed()



"""
the implementations_stack manages implementation picking
and fallback preferences

based on the contexts pushed/poped from the stack it will choose
context roots and help picking implementations
"""
from contextlib import contextmanager
from collections import namedtuple

LIMIT = 20

ImplementationChoice = namedtuple("ImplementationChoice", "key, value")


class Chooser(namedtuple("Chooser", "elements, previous, frozen")):
    @classmethod
    def make(cls, current, elements, frozen):
        if current is not None and current.frozen:
            raise RuntimeError(
                "further nesting of implementation choice has been disabled"
            )
        if frozen:
            assert len(elements) == 1

        return cls(elements, current, frozen)

    def choose(self, choose_from):
        """given a mapping of implementations
        choose one based on the current settings
        returns a key value pair
        """

        for choice in self.elements:
            if choice in choose_from:
                return ImplementationChoice(choice, choose_from[choice])
        raise LookupError(self.elements, choose_from.keys())


class NullChooser:
    frozen = False

    def choose(self, *_, **__):
        raise LookupError("No choice possible without valid context")


def chain(element):
    elements = []
    while not isinstance(element, NullChooser):
        elements.append(element)
        element = element.previous
    elements.reverse()
    return elements


class ChooserStack:
    def __init__(self, default_elements=None):
        if default_elements is not None:
            self.current = Chooser(
                elements=default_elements, previous=NullChooser(), frozen=False
            )
        else:
            self.current = NullChooser()

    def __repr__(self):
        return f"<ICS {chain(self.current)}>"

    def choose(self, choose_from):
        """given a mapping of implementations
        choose one based on the current settings
        returns a key value pair
        """
        return self.current.choose(choose_from)

    @contextmanager
    def pushed(self, new, frozen=False):
        self.current = Chooser.make(self.current, new, frozen)
        try:
            if len(chain(self.current)) > LIMIT:
                raise OverflowError("stack depth exceeded")
            yield
        finally:
            self.current = self.current.previous



import contextlib
import attr
import dectate
from collections import defaultdict
from .chooser import ChooserStack

METHOD_DATA_KEY = "sentaku_method_data"


@contextlib.contextmanager
def _use_maybe_strict(ctx, impl):
    with ctx.use(impl, frozen=ctx.strict_calls) as impl:
        yield impl


@attr.s
class ImplementationRegistrationAction(dectate.Action):
    config = {"methods": lambda: defaultdict(dict)}
    method = attr.ib()
    implementation = attr.ib()

    def identifier(self, methods):
        return self.method, self.implementation

    def perform(self, obj, methods):
        methods[self.method][self.implementation] = obj


@attr.s(hash=False)
class ImplementationContext(dectate.App):
    """maintains a mapping
    of :ref:`implementation-identification` to implementations,
    as well as the list of currently availiable Implementations
    in the order of precedence.

    :type implementations: `collections.Mapping`
    :param implementations:
        the implementations availiable in the context

        a mapping of :ref:`implementation-identification` to implementation
    :param default_choices:
        the implementations that should be used by default
        in order of percedence
    :type default_choices: optional list
    """

    implementations = attr.ib()
    implementation_chooser = attr.ib(
        default=attr.Factory(ChooserStack), converter=ChooserStack
    )
    strict_calls = attr.ib(default=False)

    external_for = dectate.directive(ImplementationRegistrationAction)

    @classmethod
    def with_default_choices(cls, implementations, default_choices, **kw):
        return cls(
            implementations=implementations,
            implementation_chooser=default_choices,
            **kw
        )

    @property
    def impl(self):
        """the currently active implementation"""
        return self.implementation_chooser.choose(self.implementations).value

    def _get_implementation_for(self, key):
        self.commit()
        implementation_set = self.config.methods[key]
        return self.implementation_chooser.choose(implementation_set)

    @classmethod
    def from_instances(cls, instances, **kw):
        """utility to create the context

        by passing a ordered list of instances
        and turning them into implementations and the default choices
        """
        return cls.with_default_choices(
            implementations={type(x): x for x in instances},
            default_choices=[type(x) for x in instances],
            **kw
        )

    @property
    def context(self):
        """alias for consistence with elements"""
        return self

    root = context

    @contextlib.contextmanager
    def use(self, *implementation_types, **kw):
        """contextmanager for controlling
        the currently active/usable implementations and
        their order of percedence

        :param `implementation-identification` implementation_types:
            the implementations availiable within the context
        :keyword bool frozen: if True prevent further nesting
        """

        def _get_frozen(frozen=False):
            return frozen

        with self.implementation_chooser.pushed(
            implementation_types, frozen=_get_frozen(**kw)
        ):
            yield self.impl


@attr.s
class _ImplementationBindingMethod:
    """bound method equivalent for :class:`ImplementationCooser`

    on call it:

    * looks up the implementation
    * freezes the context
    * calls the actual implementation
    """

    instance = attr.ib()
    selector = attr.ib()

    def __call__(self, *k, **kw):
        ctx = self.instance.context
        choice, implementation = ctx._get_implementation_for(self.selector)
        bound_method = implementation.__get__(self.instance, type(self.instance))
        with _use_maybe_strict(ctx, choice):
            return bound_method(*k, **kw)


class ContextualMethod:
    """
    descriptor for implementing context sensitive methods
    and registration of their implementations


    .. code:: python

        class Example(Element):
            action = ContextualMethod()
            @action.implemented_for("db")
            def action(self):
                pass

           @action.implemented_for("test")
           def action(self):
               pass
    """

    # todo - turn into attrs class once attribute ancoring is implemented
    def __repr__(self):
        return "<ContextualMethod>"

    def external_implementation_for(self, implementation):
        return ImplementationContext.external_for(self, implementation)

    def __get__(self, instance, *_ignored):
        if instance is None:
            return self
        return _ImplementationBindingMethod(instance=instance, selector=self)


class ContextualProperty:
    # todo - turn into attrs class once attribute ancoring is implemented
    def __init__(self):
        # setter and getter currently are lookup keys
        self.setter = self, "set"
        self.getter = self, "get"

    def external_setter_implemented_for(self, implementation):
        return ImplementationContext.external_for(self.setter, implementation)

    def external_getter_implemented_for(self, implementation):
        return ImplementationContext.external_for(self.getter, implementation)

    def __set__(self, instance, value):

        ctx = instance.context
        choice, implementation = ctx._get_implementation_for(self.setter)

        bound_method = implementation.__get__(instance, type(instance))
        with _use_maybe_strict(ctx, choice):
            return bound_method(value)

    def __get__(self, instance, owner):
        if instance is None:
            return self

        ctx = instance.context
        choice, implementation = ctx._get_implementation_for(self.getter)

        bound_method = implementation.__get__(instance, type(instance))
        with _use_maybe_strict(ctx, choice):
            return bound_method()



"""

    utility classes for implementation lookup
"""
import attr


from collections.abc import Mapping


@attr.s(frozen=True)
class ImplementationName:
    """
    utility class to declare names for implementations
    """

    name = attr.ib()
    documentation = attr.ib(repr=False, default=None)


@attr.s
class AttributeBasedImplementations(Mapping):

    holder = attr.ib()
    attribute_mapping = attr.ib(validator=attr.validators.instance_of(Mapping))

    def __getitem__(self, key):
        attribute_name = self.attribute_mapping[key]
        result = getattr(self.holder, attribute_name, self)
        if result is self:
            raise LookupError(
                "{holder!r} has no attribute {name}".format(
                    holder=self.holder, name=attribute_name
                )
            )
        return result

    def __iter__(self):
        return iter(self.attribute_mapping)

    def __len__(self):
        return len(self.attribute_mapping)



#!env python
"""
  Sentaku pypi search exampple
  ~~~~~~~~~~~~~~~~~~~~~~~~~~~~

  before running::

    $ pip install selenium  requests sentaku
"""
import argparse
import contextlib
import sentaku
import requests
import attr
from selenium.webdriver import Remote
from selenium.webdriver.common.desired_capabilities import DesiredCapabilities

parser = argparse.ArgumentParser()
parser.add_argument("query")
parser.add_argument("--fast", action="store_true")


@attr.s
class FastSearch:
    get = staticmethod(requests.get)


@attr.s
class Browser:
    driver = attr.ib()

    def __getattr__(self, key):
        return getattr(self.driver, key)


class SearchContext(sentaku.ImplementationContext):
    pass


@attr.s
class Search(sentaku.Element):
    """sentaku element for really simple pypi searching"""

    base_url = attr.ib(default="https://pypi.python.org/pypi")

    search = sentaku.ContextualMethod()
    open_page = sentaku.ContextualMethod()


@SearchContext.external_for(Search.search, Browser)
def search_browser(self, text):
    """do a slow search via the website and return the first match"""
    self.impl.get(self.base_url)
    # get the search box
    search_div = self.impl.find_element_by_id("search")
    # input the search term
    search_div.send_keys(text)
    # click the search button to initiate search
    search_div.find_element_by_xpath(
        '//*[@id="content"]//div//button[contains(@type, "submit")]'
    ).click()
    # get the first element from the list of found items
    e = self.impl.find_element_by_xpath('//*[@id="content"]//li[1]/a')
    return e.get_attribute("href")


@SearchContext.external_for(Search.search, FastSearch)
def search_fast(self, text):
    """do a sloppy quick "search" via the json index"""

    resp = self.impl.get(f"{self.base_url}/{text}/json")
    return resp.json()["info"]["package_url"]


@SearchContext.external_for(Search.open_page, Browser)
def open_page(self, url):
    self.impl.get(url)


def main(search, query):
    """main function that does the search"""
    url = search.search(query)
    print(url)
    search.open_page(url)


def cli_main():
    """cli entrypoitns, sets up everything needed"""
    SearchContext.commit()
    args = parser.parse_args()
    # open up a browser
    firefox_remote = Remote("http://127.0.0.1:4444/wd/hub", DesiredCapabilities.FIREFOX)
    with contextlib.closing(firefox_remote):
        context = SearchContext.from_instances([FastSearch(), Browser(firefox_remote)])
        search = Search(parent=context)

        if args.fast:
            with context.use(FastSearch, Browser):
                main(search, args.query)
        else:
            with context.use(Browser):
                main(search, args.query)


if __name__ == "__main__":
    cli_main()



from sentaku.context import ImplementationContext, ContextualProperty, ContextualMethod
from sentaku import Element

import pytest


def test_from_instances():
    ctx = ImplementationContext.from_instances([1, "a"])
    assert ctx.implementation_chooser.current.elements == [int, str]
    assert ctx.implementations == {int: 1, str: "a"}


class LocalContext(ImplementationContext):
    pass


class LocalElement(Element):
    method = ContextualMethod()
    prop = ContextualProperty()


@LocalContext.external_for(LocalElement.method, int)
@LocalContext.external_for(LocalElement.method, str)
@LocalContext.external_for(LocalElement.prop.setter, int)
@LocalContext.external_for(LocalElement.prop.setter, str)
@LocalContext.external_for(LocalElement.prop.getter, int)
@LocalContext.external_for(LocalElement.prop.getter, str)
def method_standin(self, value=None):
    assert (
        self.context.implementation_chooser.current.frozen == self.context.strict_calls
    )


@pytest.fixture(
    params=[
        pytest.param(True, id="nest=strict"),
        pytest.param(False, id="nest=lenient"),
    ]
)
def ctx(request):
    return LocalContext.from_instances([1, "a"], strict_calls=request.param)


@pytest.fixture
def elem(ctx):
    return LocalElement(parent=ctx)


@pytest.fixture(params=(int, str))
def impl(request):
    return request.param


def test_property(ctx, elem, impl):
    with ctx.use(impl):
        elem.prop


def test_set_property(ctx, elem, impl):
    with ctx.use(impl):
        elem.prop = 1


def test_method(ctx, elem, impl):
    with ctx.use(impl):
        elem.method()



import pytest
from sentaku.chooser import ChooserStack


@pytest.fixture
def chooser():
    return ChooserStack(None)


def test_empty(chooser):
    with pytest.raises(LookupError):
        chooser.choose({3: 3})


def test_nonempty(chooser):
    with chooser.pushed(1):
        assert chooser.current.elements == 1


def test_freezing(chooser):
    with chooser.pushed([1], frozen=True):
        with pytest.raises(RuntimeError):
            with chooser.pushed([1]):
                pass


def tet_overflow(chooser):
    def nest(n=21):
        if n:
            with chooser.pushed([1]):
                nest(n=n - 1)

    with pytest.raises(OverflowError):
        nest()


def test_choose(chooser):
    with chooser.pushed([1, 2]):
        res = chooser.choose({1: "a"})
    assert res == (1, "a")


def test_choose_missing(chooser):
    with chooser.pushed([1, 2]):
        with pytest.raises(LookupError):
            chooser.choose({3: 3})



from .utils import ImplementationName, AttributeBasedImplementations

from .modeling import Element, Collection
from .context import ContextualMethod, ContextualProperty, ImplementationContext


__all__ = [
    "Element",
    "Collection",
    "ContextualMethod",
    "ContextualProperty",
    "ImplementationContext",
    "ImplementationName",
    "AttributeBasedImplementations",
]



"""
a hideous rpc interface used to demonstrate partial implementation
"""
import attr
from . import spec


@attr.s
class PseudoRpc:
    """a hideous implementation"""

    _backend = attr.ib(repr=False)

    def make_collection(self, collection):
        """creates a todo list"""
        self._backend.create_item(collection)

    def make_item(self, collection, name):
        """creates a todo list element"""
        self._backend.get_by(collection).create_item(name)

    def has_item(self, collection, name):
        """checks for the existence of a todo list element"""
        return self._backend.get_by(collection).get_by(name) is not None

    def complete_item(self, collection, name):
        """marks a todo list element as completed"""
        self._backend.get_by(collection).get_by(name).completed = True

    def clear_completed(self, collection):
        """clears the completed elements of a todo list"""
        self._backend.get_by(collection).clear_completed()


@spec.TodoAPI.external_for(spec.TodoItem.completed.setter, spec.ViaRPC)
def todo_item_set_item_completion(item, value):
    if value:
        item.impl.complete_item(item.parent.name, item.name)
    else:
        raise NotImplementedError("rpc cant undo completion")


@spec.TodoAPI.external_for(spec.TodoCollection.create_item, spec.ViaRPC)
def create_todo_item(coll, name):
    coll.impl.make_item(coll.name, name)
    return spec.TodoItem(coll, name=name)


@spec.TodoAPI.external_for(spec.TodoCollection.get_by, spec.ViaRPC)
def get_by(self, name):
    if self.impl.has_item(self.name, name):
        return spec.TodoItem(self, name=name)


@spec.TodoAPI.external_for(spec.TodoCollection.clear_completed, spec.ViaRPC)
def clear_completed(coll):
    coll.impl.clear_completed(coll.name)


@spec.TodoAPI.external_for(spec.TodoAPI.create_collection, spec.ViaRPC)
def create_collection(api, name):
    api.impl.make_collection(name)
    return spec.TodoCollection(api, name=name)



class TodoUX:
    """
    example root UX fore todo lists
    """

    def __init__(self, app):
        self.app = app

    def get_by(self, name):
        """get a todo list ux by name

        :rtype: TodoListUX
        """
        item = self.app.get_by(name)
        return TodoListUX(ux=self, controlled_list=item)

    def create_item(self, name):
        """create a new named todo list

        :rtype: TodoListUX
        """
        item = self.app.create_item(name)

        return TodoListUX(ux=self, controlled_list=item)


class TodoListUX:
    """
    example ux for single todo lists
    .. attribute:: ux

        reference to the root ux

    .. attribute:: controlled_list

        reference of the todo list implementation

    """

    def __init__(self, ux, controlled_list):
        self.ux = ux
        self.controlled_list = controlled_list

    def get_by(self, name):
        """
        find a todo list element by name
        """
        item = self.controlled_list.get_by(name)
        if item:
            return TodoElementUX(parent=self, controlled_element=item)

    def create_item(self, name):
        """
        create a new todo list item
        """
        elem = self.controlled_list.create_item(name)
        if elem:
            return TodoElementUX(parent=self, controlled_element=elem)

    def clear_completed(self):
        """
        remove all completed elements
        """
        self.controlled_list.clear_completed()


class TodoElementUX:
    """
    ux controller element for a todo list element

    .. attribute:: parent

        the controling TodoListUX

    .. attribute:: controlled_element

        the controlled TodoElement

    """

    def __init__(self, parent, controlled_element):
        self.parent = parent
        self.controlled_element = controlled_element

    @property
    def completed(self):
        """
        completion state of the controlled element
        """
        return self.controlled_element.completed

    @completed.setter
    def completed(self, value):
        self.controlled_element.completed = value



import attr


class ElementMixin:
    @property
    def context(self):
        """the context this element belongs to"""
        return self.parent.context

    @property
    def impl(self):
        """shortcut to get the currently active application implementation"""
        return self.context.impl


@attr.s
class Element(ElementMixin):
    """Base class for all application elements

    :param parent:
        controlling object of the element which is either a context
        or a surrounding element
    :type parent: :py:class:`Element` or :py:class:`ImplementationContext`
    """

    parent = attr.ib(repr=False)


@attr.s
class Collection(Element):
    """base class for collections in the application

    :todo: generic helpers for querying
    """



import attr


@attr.s
class TodoElement:
    """
    Element of a todo list
    """

    name = attr.ib()
    completed = attr.ib(default=False)


def get_by(self, name):
    """get element by name"""
    return next((item for item in self if item.name == name), None)


def create_by_name(cls, collection_name):
    def create_item(self, name):
        "create a new named %r item"
        assert self.get_by(name) is None
        item = cls(name=name)
        getattr(self, collection_name).append(item)
        return item

    create_item.__doc__ %= cls
    return create_item


@attr.s
class TodoList:
    """a named todolist"""

    name = attr.ib()
    items = attr.ib(default=attr.Factory(list), converter=list)

    def __iter__(self):
        return iter(self.items)

    get_by = get_by
    create_item = create_by_name(TodoElement, "items")

    def clear_completed(self):
        """
        removes completed elements
        """
        self.items = [i for i in self.items if not i.completed]


@attr.s
class TodoApp:
    """
    A Basic Todo List Storage

    """

    collections = attr.ib(default=attr.Factory(list), converter=list)

    def __iter__(self):
        return iter(self.collections)

    def __repr__(self):
        return f"<TodoApp {sorted(x.name for x in self.collections)!r}>"

    get_by = get_by
    create_item = create_by_name(TodoList, "collections")
